#include "glwidget.h"

#include <QOpenGLShaderProgram>
#include <QPainter>

struct VertexData {
    QVector3D coord;
    QVector2D texCoord;
};

GLWidget::GLWidget(QWidget *parent) :
    QOpenGLWidget(parent),
    m_program(0),
    m_IBO(0),
    m_vertexDataVBO(0),
    m_isRotating(true){
    startTimer( 10 );
    QFont f = font();
    f.setPointSize( 100 );
    setFont( f );
}

GLWidget::~GLWidget() {
    delete m_program;
    glDeleteBuffers(1, &m_vertexDataVBO);
    glDeleteTextures(1, &m_texture);
}

void GLWidget::initializeGL() {
    // inizializziamo le funzioni opengl
    initializeOpenGLFunctions();

    // Creiamo l'oggetto QOpenGLShaderProgram
    m_program = new QOpenGLShaderProgram(this);

    // Compiliamo lo shader dei vertici
    if (!m_program->addShaderFromSourceFile(QOpenGLShader::Vertex, ":/vertex.glsl"))
        close();

    // Compiliamo il fragment shader
    if (!m_program->addShaderFromSourceFile(QOpenGLShader::Fragment, ":/fragment.glsl"))
        close();

    // Linkiamo gli shaders
    if (!m_program->link())
        close();

    // *** VBO vertici ***
    // vettore contenete i dati associati ai vertici
    double l = 1.0 / 2.0;
    VertexData vertexData[] = {
        // faccia 0
        {QVector3D(-l, -l,  l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l, -l,  l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l,  l,  l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l,  l,  l), QVector2D(1.0f, 1.0f)},

        // faccia 1
        {QVector3D( l, -l,  l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l, -l, -l), QVector2D(1.0f, 0.0f)},
        {QVector3D( l,  l,  l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l,  l, -l), QVector2D(1.0f, 1.0f)},

        // faccia 2
        {QVector3D( l, -l, -l), QVector2D(0.0f, 0.0f)},
        {QVector3D(-l, -l, -l), QVector2D(1.0f, 0.0f)},
        {QVector3D( l,  l, -l), QVector2D(0.0f, 1.0f)},
        {QVector3D(-l,  l, -l), QVector2D(1.0f, 1.0f)},

        // faccia 3
        {QVector3D(-l, -l, -l), QVector2D(0.0f, 0.0f)},
        {QVector3D(-l, -l,  l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l,  l, -l), QVector2D(0.0f, 1.0f)},
        {QVector3D(-l,  l,  l), QVector2D(1.0f, 1.0f)},

        // faccia 4
        {QVector3D(-l, -l, -l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l, -l, -l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l, -l,  l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l, -l,  l), QVector2D(1.0f, 1.0f)},

        // faccia 5
        {QVector3D(-l,  l,  l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l,  l,  l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l,  l, -l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l,  l, -l), QVector2D(1.0f, 1.0f)}
    };
    // genera il VBO
    glGenBuffers(1, &m_vertexDataVBO);
    // attiva il VBO
    glBindBuffer(GL_ARRAY_BUFFER, m_vertexDataVBO);
    // carica nella memoria della GPU i valori contenuti in vertexData
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertexData), vertexData, GL_STATIC_DRAW);
    // disattiva il VBO
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // *** texture ***
    // generiamo il Texture Object
    glGenTextures( 1, &m_texture);
    // bind del TO
    glBindTexture(GL_TEXTURE_2D, m_texture);

    // Parametri per la gestione del TO negli shader
    // Ripete l'immagine se necessario
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    // quando l'area è più grande della texture, interpola linearmente l'originale
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    // quando l'area è più piccola della texture, interpola linearmente l'originale
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    // modifichiamo l'immagine per adattarle
    QImage img = drawImage("8");
    // carichiamo i bit che compongono l'immagine
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img.width(), img.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, img.bits() );

    // *** IBO ***
    GLushort indices[] = { 0,  1,  2,  1,  2,  3,  // Faccia 0
                           4,  5,  6,  5,  6,  7,  // Faccia 1
                           8,  9, 10,  9, 10,  11, // Faccia 2
                           12, 13, 14, 13, 14, 15, // Faccia 3
                           16, 17, 18, 17, 18, 19, // Faccia 4
                           20, 21, 22, 21, 22, 23  // Faccia 5
                         };
    glGenBuffers(1, &m_IBO);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_IBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices, GL_STATIC_DRAW);

    // Attiviamo alcuni stati
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // Localizzaziamo gli attributi negli shader
    m_vertexPos = m_program->attributeLocation("vertex");
    m_texCoordPos = m_program->attributeLocation( "texCoord" );
    m_matrixPos = m_program->uniformLocation("matrix");
    m_texturePos = m_program->uniformLocation("texture");

    // impostiamo la prospettiva
    m_matrix.perspective(60, 4.0/3.0, 0.1, 100.0);
    m_matrix.translate(0, 0, -2);
}

void GLWidget::resizeGL(int w, int h) {
    glViewport(0, 0, w, h);
}

void GLWidget::paintGL() {
    // Puliamo la la vista
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // Linkiamo gli shader
    m_program->bind();

    // Carichiamo la matrice modello/vista
    m_program->setUniformValue(m_matrixPos, m_matrix);

    // Attiviamo il VBO dei vertici
    glBindBuffer(GL_ARRAY_BUFFER, m_vertexDataVBO); // attiva il VBO
    glEnableVertexAttribArray( m_vertexPos );
    GLsizei stride = sizeof(VertexData);        // distanza tra l'inizio di un vertice e l'inizio dell'altro
    glVertexAttribPointer( m_vertexPos,
                           3,                   // numero di elementi per vertice: 3 (x,y,z)
                           GL_FLOAT,            // tipo di elemento
                           GL_FALSE,
                           stride,              // distanza inizio-inizio
                           (const void *) 0 );  // posizione dati in VertexData

    glEnableVertexAttribArray( m_texCoordPos );
    glVertexAttribPointer( m_texCoordPos,
                           2,                   // numero di elementi per vertice: 2 (x,y)
                           GL_FLOAT,            // tipo di elemento
                           GL_FALSE,
                           stride,              // distanza inizio-inizio
                           (const void *) sizeof(QVector3D) ); // posizione dati in VertexData

    // *** texture ***
    // attiviamo l'unita' 0
    glActiveTexture(GL_TEXTURE0);
    // attiviamo la texture
    glBindTexture(GL_TEXTURE_2D, m_texture);
    // impostiamo il valore nell'unita' 0
    m_program->setUniformValue( m_texturePos, 0);

    // attiviamo l'IBO
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_IBO);

    // Disegna i triangoli
    glDrawElements( GL_TRIANGLES,       // modalita
                    36,                  // numero indici
                    GL_UNSIGNED_SHORT,  // tipologia degli indici
                    0                   // offset tra gli indici
                    );

    // Rilasciamo gli shader
    m_program->release();
}

void GLWidget::timerEvent(QTimerEvent *event) {
    Q_UNUSED(event);
    if( m_isRotating ){
        // Ruotiamo di 1 grado attorno all'asse y nel sistema di riferimento dell'osservatore
        m_matrix.rotate(1.0, 0.0, 1.0, 0.0 );
        m_matrix.rotate(0.50, 1.0, 0.0, 0.0 );

        // Aggiorniamo il disegno
        update();
    }
}

void GLWidget::mousePressEvent(QMouseEvent * event){
    Q_UNUSED(event);
    m_isRotating = !m_isRotating;
}

QImage GLWidget::drawImage( const QString & txt ){
    int border = 10;
    int side = qMax( fontMetrics().width(txt) + border , fontMetrics().height()  + border );
    QPixmap pixmap( side, side );
    pixmap.fill(Qt::white);
    QPainter painter;
    painter.begin(&pixmap);
    painter.setRenderHints(QPainter::HighQualityAntialiasing | QPainter::TextAntialiasing);
    painter.setFont(font());
    painter.setPen( Qt::red );
    painter.drawText(QRect(0,0,side,side), Qt::AlignCenter, txt );
    painter.end();

    return pixmap.toImage().convertToFormat( QImage::Format_RGBA8888 );
}
